import logging
from typing import Optional, List, Dict, Any
from datetime import datetime, timedelta

import requests
import backoff  # type: ignore[import]
import logzero  # type: ignore[import]

BASE_LISTENBRAINZ_URL = "https://api.listenbrainz.org/1/user/{username}/listens"

BASE_PLAYING_NOW_URL = "https://api.listenbrainz.org/1/user/{username}/playing-now"

Json = Any


@backoff.on_exception(
    lambda: backoff.constant(interval=10),
    exception=requests.RequestException,
    max_tries=3,
)
def request_playing_now(
    username: str,
    *,
    logger: Optional[logging.Logger] = None,
) -> List[Json]:
    """
    request the currently playing song, if any
    """
    r = requests.get(BASE_PLAYING_NOW_URL.format(username=username))
    if logger:
        logger.debug(f"Requesting {r.url}")
    r.raise_for_status()
    data = r.json()
    listens: List[Any] = data["payload"]["listens"]
    return listens


@backoff.on_exception(
    lambda: backoff.constant(interval=10),
    exception=requests.RequestException,
    max_tries=3,
)
def request_chunk(
    username: str,
    *,
    count: int = 100,
    max_ts: Optional[int] = None,
    logger: Optional[logging.Logger] = None,
) -> List[Json]:
    """
    paginate through listens for a user, by specifying an epoch time
    to receive scrobbles before
    once we receive the first chunk of scrobbles, use the epoch time of the
    last currently known scrobbles, and filter anything that was posted before that
    """
    params: Dict[str, Any] = {}
    if max_ts is not None:
        params["max_ts"] = max_ts
    params["count"] = count
    r = requests.get(BASE_LISTENBRAINZ_URL.format(username=username), params=params)
    if logger:
        logger.debug(f"Requesting {r.url}")
    r.raise_for_status()
    data = r.json()
    listens: List[Json] = data["payload"]["listens"]
    return listens


def request_listens(
        username: str,
        logger: logging.Logger = logzero.logger,
        pages: Optional[int] = None,
        days: Optional[int] = None
) -> Json:
    max_ts: Optional[int] = None
    all_listens: List[Json] = []
    curpage = 0
    while True:
        new_listens = request_chunk(username, max_ts=max_ts, logger=logger)
        all_listens.extend(new_listens)
        if len(new_listens) == 0:  # exhausted all paginations
            break
        max_ts = int(all_listens[-1]["listened_at"])
        logger.debug(
            f"Have {len(all_listens)}, now searching for listens before {datetime.utcfromtimestamp(max_ts)}..."
        )
        curpage += 1
        if pages is not None and curpage >= pages:
            break
        if days is not None and (datetime.utcfromtimestamp(max_ts) < datetime.utcnow() - timedelta(days=days)):
            break
    return all_listens



import json
from datetime import datetime, timezone
from typing import Optional, NamedTuple, Any, Dict, List, Iterator, Union

DATE_REGEX = "%a, %d %b %Y %H:%M:%S %Z"

Json = Dict[str, Any]


def _parse_date(s: Union[str, int, None]) -> Optional[datetime]:
    if s is None:
        return None
    if isinstance(s, int):
        return datetime.fromtimestamp(s, tz=timezone.utc)
    elif isinstance(s, str):
        return datetime.strptime(s, DATE_REGEX)
    else:
        raise ValueError(f"Not sure how to parse date {s} {type(s)}")


class Listen(NamedTuple):
    track_name: str
    artist_name: str
    # could be null if you're currently listening to something
    listened_at: Optional[datetime]
    inserted_at: Optional[datetime]
    recording_id: Optional[str]
    release_name: Optional[str]
    metadata: Json
    username: Optional[str]

    @classmethod
    def from_blob(cls, blob: Dict[str, Any]) -> "Listen":
        # this works by using 'pop's on the dictionary to slowly
        # decompose them without running into possible errors,
        # and taking any remaining data left over in the typical
        # structure and merging them into the 'metadata' blob,
        # since theres so many possible keys on this object
        #
        # just extracts the most useful stuff and attaches the
        # rest onto metadata
        track_metadata = blob.pop("track_metadata", {})
        listened_at: Optional[int] = blob.pop("listened_at", None)
        inserted_at: Optional[str] = blob.pop("inserted_at", None)
        recording_id: Optional[str] = blob.pop("recording_msid", None)
        username: Optional[str] = blob.pop("user_name", None)
        release_name = track_metadata.pop("release_name", None)
        artist_name: str = track_metadata.pop("artist_name", "<unknown>")
        track_name: str = track_metadata.pop("track_name", "<unknown>")
        # merge all other additional info into the JSON object
        metadata: Json = {}
        additional_info = track_metadata.pop("additional_info", {})
        metadata.update(additional_info)
        metadata.update(track_metadata)
        metadata.update(blob)
        return cls(
            track_name=track_name,
            artist_name=artist_name,
            listened_at=_parse_date(listened_at),
            inserted_at=_parse_date(inserted_at),
            recording_id=recording_id,
            release_name=release_name,
            metadata=metadata,
            username=username,
        )


def iter_listens(from_file: str) -> Iterator[Listen]:
    with open(from_file, "r") as f:
        data: List[Dict[str, Any]] = json.load(f)
    for blob in data:
        yield Listen.from_blob(blob)



import json
from typing import Optional

import click

from .export import request_listens, request_playing_now


@click.group()
def main() -> None:
    pass


@main.command()
@click.argument("LISTENBRAINZ_USERNAME")
def playing_now(listenbrainz_username: str) -> None:
    """
    Downloads your currently playing track from listenbrainz
    """
    data = request_playing_now(username=listenbrainz_username)
    click.echo(json.dumps(data))
    exit(0 if len(data) else 1)


@main.command()
@click.option(
    "-p",
    "--pages",
    type=int,
    default=None,
    help="Request these many pages of your history",
)
@click.option(
    "--days",
    type=int,
    default=None,
    help="Request listens no older than this many days",
)
@click.argument("LISTENBRAINZ_USERNAME")
def export(pages: Optional[int], days: Optional[int], listenbrainz_username: str) -> None:
    """
    Downloads all scrobbles for your listenbrainz account
    """
    click.echo(json.dumps(request_listens(username=listenbrainz_username, pages=pages, days=days)))


if __name__ == "__main__":
    main(prog_name="listenbrainz_export")
